// Copyright 2018 The gVisor Authors.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

// +build !linux

// Package rand implements a cryptographically secure pseudorandom number
// generator.
package rand

import "crypto/rand"

// Reader is the default reader.
var Reader = rand.Reader

// Read implements io.Reader.Read.
func Read(b []byte) (int, error) {
	return rand.Read(b)
}
